import random
from functools import reduce
import os
import torch
import numpy as np
import matplotlib
import torchvision
import torch.nn.functional
import matplotlib.pyplot as plt
import torch.utils.data
import shutil

MAX_LEN = 200 # For debugging, reduce number of samples

BATCH_SIZE = 18
ANCHOR_SIZE = int(BATCH_SIZE/3)
BATCH_SIZE_POOL = 64

LEARNING_RATE = 1e-4
IS_HEADLESS = True
RUN_NAME = 'test_contrastive_4'
MARGIN = 0.2
EMBEDDING_SIZE = 32

DEVICE = 'cpu'
if torch.cuda.is_available():
    DEVICE = 'cuda'
    MAX_LEN = 0

print(f'training on: {DEVICE}')

path_run = f'./results/{RUN_NAME}'
if os.path.exists(path_run):
    shutil.rmtree(path_run)
os.makedirs(path_run)

class DatasetFassionMNIST(torch.utils.data.Dataset):
    def __init__(self, is_train):
        super().__init__()
        self.data = torchvision.datasets.FashionMNIST(
            root='../data',
            train=is_train,
            download=True
        )

    def __len__(self):
        if MAX_LEN:
            return MAX_LEN
        return len(self.data)

    def __getitem__(self, idx):
        # list tuple np.array torch.FloatTensor
        pil_x, y_idx = self.data[idx]
        np_x = np.array(pil_x)
        np_x = np.expand_dims(np_x, axis=0) # (1, W, H)

        x = torch.FloatTensor(np_x)
        y = torch.LongTensor([y_idx])
        return x, y


data_loader_train = torch.utils.data.DataLoader(
    dataset=DatasetFassionMNIST(is_train=True),
    batch_size=BATCH_SIZE_POOL,
    shuffle=True
)

data_loader_test = torch.utils.data.DataLoader(
    dataset=DatasetFassionMNIST(is_train=False),
    batch_size=BATCH_SIZE_POOL,
    shuffle=False
)


class Model(torch.nn.Module):
    def __init__(self):
        super().__init__()

        self.encoder = torchvision.models.resnet18(
            pretrained=True
        )

        w_conv_1 = self.encoder.conv1.weight.data.mean(dim=1).unsqueeze(dim=1)
        self.encoder.conv1 = torch.nn.Conv2d(
            in_channels=1,
            out_channels=self.encoder.conv1.out_channels,
            kernel_size=7, stride=2, padding=3,
            bias=False)
        self.encoder.conv1.weight.data = w_conv_1

        self.encoder.fc = torch.nn.Linear(
            self.encoder.fc.in_features,
            EMBEDDING_SIZE
        )

    def forward(self, x):
        out = self.encoder.forward(x)
        norm = torch.norm(out.detach(), p=2, dim=1, keepdim=True)
        out = out / norm
        return out

def sample_mining(z, y):
    #TODO
    return z, y

model = Model()

model = model.to(DEVICE)
optimizer = torch.optim.Adam(model.parameters(), lr=LEARNING_RATE)

metrics = {}
for stage in ['train', 'test']:
    for metric in [
        'loss',
        'acc',
        'clust_acc',
        'dist_pos',
        'dist_neg'
    ]:
        metrics[f'{stage}_{metric}'] = []

for epoch in range(1, 100):
    plt.clf()

    for data_loader in [data_loader_train, data_loader_test]:
        metrics_epoch = {key: [] for key in metrics.keys()}

        stage = 'train'
        if data_loader == data_loader_test:
            stage = 'test'

        for x, y in data_loader:

            x = x.to(DEVICE)
            y = y.to(DEVICE).squeeze()
            z = model.forward(x)

            z, y = sample_mining(z, y)

            if len(z) == 0:
                print('no good samples to train')
                continue

            losses = []
            dist_pos = []
            dist_neg = []
            for i, y_i in enumerate(y):

                dists_all = torch.nn.functional.pairwise_distance(z[i].expand(z.size()), z, p=2)

                dists = dists_all.clone()
                dists[y == y_i] = float('Inf')
                # left non matching classes

                j = torch.argmin(dists)
                if dists[j] < float('Inf'):
                    losses.append(torch.relu(MARGIN - dists[j]))

                dists = dists_all.clone()
                dists[y != y_i] = float('-Inf')
                j = torch.argmax(dists)
                if dists[j] > float('-Inf'):
                    losses.append(dists[j])

                dist_pos.append(torch.mean(dists_all[y == y_i].detach()).cpu().item())
                dist_neg.append(torch.mean(dists_all[y != y_i].detach()).cpu().item())

            loss = torch.mean(torch.stack(losses))

            metrics_epoch[f'{stage}_loss'].append(loss.cpu().item()) # Tensor(0.1) => 0.1f

            metrics_epoch[f'{stage}_dist_pos'].append(np.mean(dist_pos))
            metrics_epoch[f'{stage}_dist_neg'].append(np.mean(dist_neg))

            if data_loader == data_loader_train:
                loss.backward()
                optimizer.step()
                optimizer.zero_grad()



        with torch.no_grad():
            class_centers = torch.zeros((len(data_loader_train.dataset.data.classes), EMBEDDING_SIZE)).to(DEVICE)
            class_centers_n = torch.zeros((len(data_loader_train.dataset.data.classes),)).to(DEVICE)
            for x, y in data_loader:
                x = x.to(DEVICE)
                y = y.to(DEVICE).squeeze()
                z = model.forward(x)

                for i, y_i in enumerate(y):
                    class_centers[y_i] += z[i]
                    class_centers_n[y_i] += 1.0

            class_centers /= class_centers_n.unsqueeze(dim=1)
            norm = torch.norm(class_centers.detach(), p=2, dim=1, keepdim=True)
            class_centers = class_centers / norm

            for x, y in data_loader:
                x = x.to(DEVICE)
                y = y.to(DEVICE).squeeze()
                z = model.forward(x)

                acc = []
                for i, y_i in enumerate(y):
                    dists = torch.nn.functional.pairwise_distance(z[i].expand(class_centers.size()), class_centers, p=2)
                    y_prim_i = torch.argmin(dists)
                    acc.append((y_prim_i == y_i) * 1.0)

                clust_acc = 0

                #TODO

                metrics_epoch[f'{stage}_clust_acc'].append(clust_acc)

                metrics_epoch[f'{stage}_acc'].append(torch.mean(torch.stack(acc)).item())


        metrics_strs = []
        for key in metrics_epoch.keys():
            if stage in key:
                value = np.mean(metrics_epoch[key])
                metrics[key].append(value)
                metrics_strs.append(f'{key}: {round(value, 2)}')

        print(f'epoch: {epoch} {" ".join(metrics_strs)}')

    plt.clf()
    plts = []
    c = 0
    for key, value in metrics.items():

        plts += plt.plot(value, f'C{c}', label=key)
        ax = plt.twinx()
        c += 1

    plt.legend(plts, [it.get_label() for it in plts])

    if IS_HEADLESS:
        plt.savefig(f'{path_run}/{epoch}.png')
    else:
        plt.show()
